SUBROUTINE &
     sub_globalegm(exp_val,exp_dval,z_node,c_node,val_node,k)
! Modified egm algorithm. Solves for consumption, value fn,
!       endog. cash-at-hand nodes and their number
! Inputs: 
! 	exp_val  (DISCOUNTED expected value fn as a function of wealth ONLY )
!	exp_dval (partial derivative of exp_val with respect to wealth)
! Outputs:
!	c_node, val_node (consumption and value fn on the endogenous grid) 
!	z_node (associated endogenous grid point for cash-at-hand)
!	k (number of nodes for the policy and value fn)
!
! USER: the algorithm uses the non-linear solver in sub_newton to allow 
!       for non-invertible marginal felicity.
!       If your marginal felicity function is invertible you may replace 
!       the lines 
!             b_global = exp_dval(index_a)
!             CALL sub_newton(sub_FCN1,sub_D2FCN,xguess,x)
!       with 
!             x = u'^(-1)(exp_dval(index_a))
!       where u'^(-1) is the inverse of the marginal felicity fn.
!       In my experience, this makes hardly any difference in terms 
!       of speed.
 
  USE mod_types
  USE mod_common_var  
  USE mod_felicity  
  USE mod_fcn        
  USE qsort_c_module
  USE mod_solver

  IMPLICIT NONE

  REAL (dp), INTENT(in),  DIMENSION(len_a_grid)	:: exp_val, exp_dval
  REAL (dp), INTENT(out), DIMENSION(len_a_grid) :: c_node, val_node, z_node
  INTEGER, INTENT(out) :: k
  
  LOGICAL, DIMENSION(SIZE(exp_val)-1) :: mask

  INTEGER  :: i, i_0, index_a, index_a_min, index_a_max, j, j_0, max_loc
  INTEGER, DIMENSION(len_a_grid) :: aprime_node, i_aprime

  REAL(dp) :: exp_dval_down, exp_dval_up, val_temp, z_bar, z_temp
  REAL(dp), DIMENSION(len_a_grid) :: c_local, z_local
  REAL(dp), DIMENSION(2) :: c_temp, val

  
  !Variables used by sub_newton subroutine
  INTEGER, PARAMETER    :: N=1
  REAL(dp)              :: FNORM, F(N), X(N), XGUESS(N)
    
    
  ! Initialize indices of the bounds of the concave region.
  index_a_min = len_a_grid
  index_a_max = 1
  
  ! 1. Check for non-concavity and compute bounds of non-concave region
  !-------------------------------------------------------------------
  ! Find discontinuities (upward jumps) in future marginal utility
  mask = (exp_dval(2:len_a_grid)>exp_dval(1:len_a_grid-1))
  IF (COUNT(mask)>0) THEN 
     ! Non-concavity. Computes bounds of region.
     exp_dval_up = MAXVAL(PACK(exp_dval(2:len_a_grid),mask))
     exp_dval_down = MINVAL(PACK(exp_dval(1:len_a_grid-1),mask))
     IF (COUNT(exp_dval>exp_dval_up)==0) THEN 
        index_a_min=1
     ELSE
        index_a_min = MINLOC(exp_dval,1,exp_dval>exp_dval_up)+1
     END IF
     IF (COUNT(exp_dval<exp_dval_down)==0) THEN 
        index_a_max=len_a_grid 
     ELSE
        index_a_max = MAXLOC(exp_dval,1,exp_dval<exp_dval_down) - 1
     END IF
  END IF
 
  ! 2. Solve for endogenous grid points
  !-------------------------------------------------------------------
  ! Use EGM to solve for local maxima
  xguess = .1d0
  DO index_a = 1,len_a_grid
    b_global = exp_dval(index_a)
    CALL sub_newton(sub_FCN1,sub_D2FCN,xguess,x,0.d0)
    c_local(index_a) = x(1)
    z_local(index_a) = x(1) + a_grid(index_a)
    xguess = x
  ENDDO
  z_bar = z_local(1)
    
  i_aprime = (/ (i,i=1,len_a_grid) /)
  IF ( index_a_min<len_a_grid .AND. index_a_max>1 ) THEN
    !If exp_dval is discontinuous, sort local maxima increasingly in z
    ! and permute their index accordingly
    CALL QsortC(z_local(index_a_min:index_a_max), &
      i_aprime(index_a_min:index_a_max))
    c_local(index_a_min:index_a_max) = &
      c_local(i_aprime(index_a_min:index_a_max))
  ENDIF
 
  i_0 = 0
  j_0 = 0
  k = 0
  DO i= 1,len_a_grid
    IF (i<index_a_min .OR. i>index_a_max) THEN  !if_1
      ! 2.1 Concave region - local max are global
      ! ----------------------------------
      k = k+1
      c_node(k) = c_local(i)
      z_node(k) = z_local(i)
      aprime_node(k) = i_aprime(i)
      i_0 = i_aprime(i)
      j_0 = i
      val_node(k) = fel(c_node(k)) + exp_val(i_aprime(i))
    ! 2.2 Non-concave region - discard local but not global max
    ! ----------------------------------
    ELSE IF (k>0 .AND. i_aprime(i)<i_0 ) THEN
      CONTINUE
    ELSE    
      val(2) = fel(z_local(i)-a_grid(i_aprime(i))) + exp_val(i_aprime(i))
      IF (k==0) j_0 = i
      j = MIN(i+10,index_a_max)
        val(1) = MAXVAL( fel(z_local(i)-a_grid(i_aprime(j_0:j)))  + & 
        exp_val(i_aprime(j_0:j)) )
        IF ( val(1)>val(2) ) CYCLE
        IF ( z_local(i)<z_bar ) THEN
         val(1) = fel(z_local(i)-a_grid(1)) + exp_val(1)
            IF ( val(1)>val(2) ) CYCLE
          ENDIF
          k = k+1
          c_node(k) = c_local(i)
          z_node(k) = z_local(i)
          aprime_node(k) = i_aprime(i)
          i_0 = i_aprime(i)
          j_0 = i
          val_node(k) = fel(c_node(k)) + exp_val(i_aprime(i))
        
   ENDIF
  ENDDO
   
      
  !3. Correct the policy and value fns on the first end. cash-at-hand
  !   node if the choice of future assets is to the right of the
  !   borrowing constraint
  !----------------------------------------------------------
  
  IF (aprime_node(1)>1) THEN
    b_global = a_grid(aprime_node(1))
    fut_v1_global = exp_val(aprime_node(1)) - exp_val(1)
    xguess = z_node(1)
    ! Solve for value of cash at hand to the left of which it is 
    ! optimal to be borrowing constrained and for the associated values
    ! of the consumption and value fn
    CALL sub_newton(sub_min_z,sub_dmin_z,xguess,x,b_global)
    c_node(2:len_a_grid) = c_node(1:len_a_grid-1)
    z_node(2:len_a_grid) = z_node(1:len_a_grid-1)
    val_node(2:len_a_grid) = val_node(1:len_a_grid-1)
    c_node(1) = -1.d-5 + x(1) 
    z_node(1) = c_node(1) + a_grid(1)
    val_temp = &
        fel(c_node(1))+ exp_val(1)
    val_node(1) = val_temp
    k = k + 1 
  END IF

END SUBROUTINE sub_globalegm
